# -*- coding: utf-8 -*-
#
# New BSD license
#
# Copyright (c) DR0ID
# This file is part of HG_pyknic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL DR0ID BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This module contains generators to generate ids or flags.
"""
from __future__ import print_function, division

import logging
logger = logging.getLogger(__name__)
logger.debug("importing...")

__version__ = '1.0.0.0'

# for easy comparison as in sys.version_info but digits only
__version_info__ = tuple([int(d) for d in __version__.split('.')])

__author__ = "DR0ID"
__email__ = "dr0iddr0id {at} gmail [dot] com"
__copyright__ = "DR0ID @ 2015"
__credits__ = ["DR0ID"]  # list of contributors
__maintainer__ = "DR0ID"
__license__ = "New BSD license"

__all__ = ["IdGenerator", "GlobalIdGenerator", "FlagGenerator"]  # list of public visible parts of this module


class IdGenerator(object):

    """
    An Id generator.
    :param initial_value: The initial value to use.
    """

    def __init__(self, initial_value):
        self._current = initial_value
        logger.debug(self.__class__.__name__ + " created " + str(id(self)))

    def _foo(self):
        self._current += 1
        yield self._current

    def next(self):
        gen = self._foo()
        while True:
            return next(gen)

    @property
    def last_id(self):
        """
        Returns the last generated id.
        :return: last generated id
        """
        return self._current

    def set_last_id(self, last_id):
        """
        Set the last id that has been used. It is only stored if its greater than the previous id.
        :param last_id: The last used id.
        :return: None
        """
        if last_id > self._current:
            self._current = last_id


class GlobalIdGenerator(object):

    """
    A global id generator, once initialized it will generate an unique id independent of how many instance are created.
    :param initial_value: The initial value of the first instance created will be used, all other are ignored!
    """
    _id_generator = None

    def __init__(self, initial_value=0):
        if GlobalIdGenerator._id_generator is None:
            GlobalIdGenerator._id_generator = IdGenerator(initial_value)
        logger.debug(self.__class__.__name__ + " created " + str(id(self)))

    def next(self):
        return GlobalIdGenerator._id_generator.next()


class FlagGenerator(object):
    """
    This is a special kind of generator. It creates bit flags. E.g. 1, 2, 4, 8, ... basically doing
    1 << x where x increases with each call.
    """

    def __init__(self):
        self._id_generator = IdGenerator(-1)
        logger.debug(self.__class__.__name__ + " created " + str(id(self)))

    def next(self):
        """Returns the next bit flag."""
        return 1 << self._id_generator.next()

logger.debug("imported")
